"""
HCF Backend for Frontera Scheduler

Optimizing frontier setting configuration:
------------------------------------------

If you want to limit each consumer job, use one of the
following parameters. One limits by read requests count,
the other by read batches count. In HCF each batch contains
no more than 100 requests.
HCF_CONSUMER_MAX_REQUESTS = 15000
HCF_CONSUMER_MAX_BATCHES = 150

How many requests are read on each call to HCF. If 0, limit
is determined by the speed at which scrapy demands requests, which
is very unnefficient because it makes lots of calls to HCF which retrieves
a small number of requests. Same if you give it a small value. A too big value
will make to retrieve a total of requests/batches bigger than setted
in HCF_CONSUMER_MAX_* settings. Can also make to fail requests to HCF
by timeout. A value of 1000 is optimal for most purposes.
MAX_NEXT_REQUESTS = 1000

Another important setting that affects performance is HCF_PRODUCER_BATCH_SIZE. This is the number
of requests accumulated per slot before syncing to the frontier. Hence, this setting affects
the memory needed for the producer. The default of 1000 is good enough for most purposes. But if
you still have memory issues, specially when using many slots, you may want to decrease it.

Usage details
-------------

In order to enable, follow instructions on how to enable crawl frontier scheduler on frontera doc, and set
frontier BACKEND setting to hcf_backend.HCFBackend. Simple example::

in mycrawler/settings.py::

SCHEDULER = 'frontera.contrib.scrapy.schedulers.frontier.FronteraScheduler'
FRONTERA_SETTINGS = 'mycrawler.frontera_settings'

in mycrawler/frontier_settings.py::

BACKEND = 'hcf_backend.HCFBackend'
HCF_CONSUMER_MAX_BATCHES = 500
MAX_NEXT_REQUESTS = 1000

Backend settings:
-----------------

* HCF_AUTH - Hubstorage auth (not required if SH_APIKEY environment variable is set up)
* HCF_PROJECT_ID - Hubstorage project id (not required if job run in scrapinghub ScrapyCloud, or PROJECT_ID
                   environment variable is set up)

If is producer:
* HCF_PRODUCER_FRONTIER - The frontier where URLs are written.
* HCF_PRODUCER_SLOT_PREFIX - Prefix to use for slot names.
* HCF_PRODUCER_NUMBER_OF_SLOTS - Number of write slots to use.
* HCF_PRODUCER_BATCH_SIZE - How often slot flush should be called. When a slot reaches the number, it is flushed.
  HCF max batch size is 100, so if HCF_PRODUCER_BATCH_SIZE is greater than 100, will result on more than one HCF batches
  of up to 100 requests each. If HCF_PRODUCER_BATCH_SIZE is lower than 100, will result on batches up to the given size.


If is consumer:
* HCF_CONSUMER_FRONTIER - The frontier where URLs are readed.
* HCF_CONSUMER_SLOT - Slot from where the spider will read new URLs.
* HCF_CONSUMER_MAX_BATCHES - Max batches to read from hubstorage.
* HCF_CONSUMER_MAX_REQUESTS - Max request to be read from hubstorage.
    (note: crawler stops to read from hcf when any of max batches or max requests limit are reached)
* HCF_CONSUMER_DONT_DELETE_REQUESTS - If given and True, don't delete requests from frontier once read.
  For testing purposes or delegation of batch control.
* HCF_CONSUMER_DELETE_BATCHES_ON_STOP - If given and True, read batches will be deleted when the job finishes.
    Default is to delete batches once read.

"""

import datetime
import logging

from frontera import Backend
from shub_workflow.utils import resolve_project_id

from hcf_backend.manager import HCFManager
from hcf_backend.utils import (
    convert_from_bytes,
    convert_to_bytes,
    assign_slotno,
)


__all__ = ["HCFBackend"]

LOG = logging.getLogger(__name__)


DEFAULT_HCF_PRODUCER_NUMBER_OF_SLOTS = 1
DEFAULT_HCF_PRODUCER_SLOT_PREFIX = ""
DEFAULT_HCF_PRODUCER_BATCH_SIZE = 1000
DEFAULT_HCF_CONSUMER_SLOT = "0"
DEFAULT_HCF_CONSUMER_MAX_BATCHES = 0
DEFAULT_HCF_CONSUMER_MAX_REQUESTS = 0


class HCFBackend(Backend):

    backend_settings = (
        "HCF_AUTH",
        "HCF_PROJECT_ID",
        "HCF_PRODUCER_FRONTIER",
        "HCF_PRODUCER_SLOT_PREFIX",
        "HCF_PRODUCER_NUMBER_OF_SLOTS",
        "HCF_PRODUCER_BATCH_SIZE",
        "HCF_CONSUMER_FRONTIER",
        "HCF_CONSUMER_SLOT",
        "HCF_CONSUMER_MAX_BATCHES",
        "HCF_CONSUMER_MAX_REQUESTS",
        "HCF_CONSUMER_DONT_DELETE_REQUESTS",
        "HCF_CONSUMER_DELETE_BATCHES_ON_STOP",
    )

    component_name = "HCF Backend"

    @classmethod
    def from_manager(cls, manager):
        return cls(manager)

    def __init__(self, manager):
        self.manager = manager

        self.hcf_auth = None
        self.hcf_project_id = resolve_project_id()

        self.hcf_producer_frontier = None
        self.hcf_producer_slot_prefix = DEFAULT_HCF_PRODUCER_SLOT_PREFIX
        self.hcf_producer_number_of_slots = DEFAULT_HCF_PRODUCER_NUMBER_OF_SLOTS
        self.hcf_producer_batch_size = DEFAULT_HCF_PRODUCER_BATCH_SIZE

        self.hcf_consumer_frontier = None
        self.hcf_consumer_slot = DEFAULT_HCF_CONSUMER_SLOT
        self.hcf_consumer_max_batches = DEFAULT_HCF_CONSUMER_MAX_BATCHES
        self.hcf_consumer_max_requests = DEFAULT_HCF_CONSUMER_MAX_REQUESTS
        self.hcf_consumer_dont_delete_requests = False
        self.hcf_consumer_delete_batches_on_stop = False

        self.stats = self.manager.settings.get("STATS_MANAGER")

        self.n_consumed_batches = 0
        self.n_consumed_requests = 0

        self.producer = None
        self.consumer = None

        self.consumed_batches_ids = []
        self._no_last_data = False

    def frontier_start(self):
        for attr in self.backend_settings:
            value = self.manager.settings.get(attr)
            if value is not None:
                setattr(self, attr.lower(), value)
        try:
            int(self.hcf_project_id)
        except TypeError:
            LOG.warning("Could not detect project. You must set HCF_PROJECT_ID setting. HCFBackend is not configured.")
        else:
            self._init_roles()
            self._log_start_message()

    def _get_producer_newcounts(self):
        try:
            newcount = self.producer._hcf.newcount
        except Exception:
            pass
        else:
            yield None, newcount

        # TODO: add per-slot newcounts in python-scrapinghub & here.

    def _update_producer_new_links_stat(self):
        for slot, newcount in self._get_producer_newcounts():
            self.stats.set_value(self._get_producer_stats_msg(slot=slot, msg="new_links"), newcount)

    def frontier_stop(self):
        if self.producer:
            self.producer.flush()
            self.producer.close()
            self._update_producer_new_links_stat()

        if self.consumer:
            if not self.hcf_consumer_dont_delete_requests:
                self.delete_read_batches()

            self.consumer.close()

    def add_seeds(self, seeds):
        for request in seeds:
            self._process_hcf_link(request)
        self._update_producer_new_links_stat()

    def page_crawled(self, response):
        pass

    def links_extracted(self, request, links):
        for req in links:
            self._process_hcf_link(req)
        self._update_producer_new_links_stat()

    def get_next_requests(self, max_next_requests, **kwargs):

        requests = []

        if self.hcf_consumer_max_requests > 0:
            max_next_requests = min(max_next_requests, self.hcf_consumer_max_requests - self.n_consumed_requests)

        if (
            self.consumer
            and not (self._consumer_max_batches_reached() or self._consumer_max_requests_reached())
            and max_next_requests
        ):
            for request in self._get_requests_from_hs(max_next_requests):
                requests.append(request)

        return requests

    @staticmethod
    def _convert_qdata_to_bytes(qdata):
        req = qdata["request"]
        req["headers"] = convert_to_bytes(req.get("headers", {}))
        req["cookies"] = convert_to_bytes(req.get("cookies", {}))
        for key in list(req.get("meta", {}).keys()):
            req["meta"][key.encode("utf8")] = req["meta"].pop(key)

    def _get_requests_from_hs(self, n_min_requests):
        return_requests = []
        data = True

        if self.producer is not None:
            self.producer.flush()

        while (
            data
            and len(return_requests) < n_min_requests
            and not (self._consumer_max_batches_reached() or self._consumer_max_requests_reached())
        ):
            data = False
            for batch in self.consumer.read(self.hcf_consumer_slot, n_min_requests):
                data = True
                batch_id = batch["id"]
                if batch_id in self.consumed_batches_ids:
                    return return_requests
                requests = batch["requests"]
                self.stats.inc_value(self._get_consumer_stats_msg("requests"), len(requests))
                for fingerprint, qdata in requests:
                    self._convert_qdata_to_bytes(qdata)
                    request = self._make_request(fingerprint, qdata)
                    if request is not None:
                        request.meta.update(
                            {b"created_at": datetime.datetime.utcnow(), b"depth": 0}
                        )
                        return_requests.append(request)
                        self.n_consumed_requests += 1
                self.consumed_batches_ids.append(batch_id)
                self.n_consumed_batches += 1
                self.stats.inc_value(self._get_consumer_stats_msg("batches"))
                LOG.info("Reading %d request(s) from batch %s ", len(requests), batch_id)

            if not self.hcf_consumer_dont_delete_requests and not self.hcf_consumer_delete_batches_on_stop:
                self.delete_read_batches()

        self._no_last_data = not data
        return return_requests

    def delete_read_batches(self):
        if self.consumed_batches_ids:
            self.consumer.delete(self.hcf_consumer_slot, self.consumed_batches_ids)
            LOG.info("Deleting read batches: %s", self.consumed_batches_ids)
        self.consumed_batches_ids = []

    def _make_request(self, fingerprint, qdata):
        kwargs = qdata["request"]
        kwargs["meta"][b"frontier_fingerprint"] = fingerprint
        return self.manager.request_model(qdata["url"], **kwargs)

    def _log_start_message(self):
        producer_message = "NO"
        consumer_message = "NO"
        if self.producer:
            if self.hcf_producer_number_of_slots > 1:
                slots_message = "[0-%d]" % (self.hcf_producer_number_of_slots - 1)
            else:
                slots_message = "0"
            producer_message = "%s/%s%s" % (self.hcf_producer_frontier, self.hcf_producer_slot_prefix, slots_message)
        if self.consumer:
            consumer_message = "%s/%s" % (self.hcf_consumer_frontier, self.hcf_consumer_slot)
        LOG.info("HCF project: %s", self.hcf_project_id)
        LOG.info("HCF producer: %s", producer_message)
        LOG.info("HCF consumer: %s", consumer_message)

    def _process_hcf_link(self, link):
        if not self.producer:
            LOG.debug(f"Ignoring {link.url}: backend not configured as producer")
            return
        link.meta.pop(b"origin_is_frontier", None)
        hcf_request = {"fp": link.meta[b"frontier_fingerprint"]}
        qdata = {"request": {}}
        for attr in ("method", "headers", "cookies", "meta"):
            qdata["request"][attr] = getattr(link, attr)
        qdata["url"] = link.url
        hcf_request["qdata"] = qdata

        slot = self.hcf_get_producer_slot(link)
        if slot:
            self.producer.add_request(slot, convert_from_bytes(hcf_request))

            self.stats.inc_value(self._get_producer_stats_msg(slot, msg="total_links"))
            self.stats.inc_value(self._get_producer_stats_msg(msg="total_links"))

    def _consumer_max_batches_reached(self):
        if not self.hcf_consumer_max_batches:
            return False
        return self.n_consumed_batches >= self.hcf_consumer_max_batches

    def _consumer_max_requests_reached(self):
        if not self.hcf_consumer_max_requests:
            return False
        return self.n_consumed_requests >= self.hcf_consumer_max_requests

    def _init_roles(self):

        if self.hcf_producer_frontier:
            self.producer = HCFManager(
                project_id=self.hcf_project_id,
                frontier=self.hcf_producer_frontier,
                auth=self.hcf_auth,
                batch_size=self.hcf_producer_batch_size,
            )

        if self.hcf_consumer_frontier and self.hcf_consumer_slot:
            self.consumer = HCFManager(
                project_id=self.hcf_project_id, frontier=self.hcf_consumer_frontier, auth=self.hcf_auth
            )

    def hcf_get_producer_slot(self, request):
        """Determine to which slot should be saved the request.

        Depending on the urls, this distribution might or not be evenly among
        the slots.
        """
        fingerprint = request.meta[b"frontier_fingerprint"]
        slot_prefix = request.meta.get(b"frontier_slot_prefix", self.hcf_producer_slot_prefix)
        num_slots = request.meta.get(b"frontier_number_of_slots", self.hcf_producer_number_of_slots)
        slotno = assign_slotno(fingerprint, num_slots)
        slot = slot_prefix + str(slotno)
        return slot

    def _get_consumer_stats_msg(self, msg=None):
        stats_msg = "hcf/consumer/%s/%s" % (self.hcf_consumer_frontier, self.hcf_consumer_slot)
        if msg:
            stats_msg += "/%s" % msg
        return stats_msg

    def _get_producer_stats_msg(self, slot=None, msg=None):
        stats_msg = "hcf/producer/%s" % (self.hcf_producer_frontier)
        if slot:
            stats_msg += "/%s" % slot
        if msg:
            stats_msg += "/%s" % msg
        return stats_msg

    def request_error(self, request, error):
        pass

    def finished(self):
        return self._no_last_data and (self.producer is None or self.producer.get_number_of_links_to_flush() == 0)



import re
import pprint
from itertools import cycle, groupby
from operator import itemgetter

import humanize
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError, ConnectionError
from retrying import retry
from shub_workflow.script import BaseScript

from hcf_backend.utils import assign_slotno


def retry_if_http_error(e):
    for exctype in HTTPError, ConnectionError:
        if isinstance(e, exctype):
            print(f"Error: {e}. Retrying...")
            return True
    return False


class HCFPal:

    HCF_API_URLS = {
        "count": "https://storage.scrapinghub.com/hcf/{pid}/{frontier}/s/{slot}/q/count",
        "list_frontiers": "https://storage.scrapinghub.com/hcf/{pid}/list",
        "list_slots": "https://storage.scrapinghub.com/hcf/{pid}/{frontier}/list",
    }

    def __init__(self, project):
        self.project = project

    @property
    def projectid(self):
        return self.project.projectid

    @property
    def auth(self):
        return self.project.auth

    @retry(
        retry_on_exception=retry_if_http_error, wait_fixed=60000, stop_max_attempt_number=60 * 24,  # 1 min
    )  # 1 day
    def _get_json(self, url):
        response = requests.get(url, auth=HTTPBasicAuth(*self.auth))
        response.raise_for_status()
        return response.json()

    def get_frontiers(self):
        url = self.HCF_API_URLS["list_frontiers"].format(pid=self.projectid)
        return self._get_json(url)

    def get_slots(self, frontier):
        url = self.HCF_API_URLS["list_slots"].format(pid=self.projectid, frontier=frontier)
        return self._get_json(url)

    def delete_slots(self, frontier, slots):
        for slot in slots:
            self.project.frontier.delete_slot(frontier, slot)

    def list_all(self, prettyprint=False):
        yall = {}
        for frontier in self.get_frontiers():
            yall[frontier] = self.get_slots(frontier)
        if prettyprint:
            return pprint.pformat(yall, indent=4)
        else:
            return yall

    def get_slot_count(self, frontier, slot):
        URL = self.HCF_API_URLS["count"].format(pid=self.projectid, frontier=frontier, slot=slot)
        total = 0
        nextstart = ""
        while True:
            next_url = URL + "?start={}".format(nextstart) if nextstart else URL
            data = self._get_json(next_url)
            total += int(data.get("count"))
            nextstart = data.get("nextstart", "")
            if not nextstart:
                break
        return total

    def get_slots_count(self, frontier, prefix, num_slots=None, regex=""):
        result = {"slots": {}}
        total = 0
        not_empty_slots = 0
        slots = ["{}{}".format(prefix, slot) for slot in range(num_slots)] if num_slots else self.get_slots(frontier)
        for slot in slots:
            if not slot.startswith(prefix):
                continue
            if not re.search(regex, slot):
                continue
            cnt = self.get_slot_count(frontier, slot)
            if cnt:
                not_empty_slots += 1
            total += cnt
            result["slots"][slot] = cnt
        result["total"] = total
        result["not empty slots"] = not_empty_slots
        return result

    def dump_slot(self, frontier, slot, max_requests):
        count = 0
        for batch in self.project.frontier.read(frontier, slot, max_requests):
            for request in batch["requests"]:
                yield batch["id"], request
                count += 1
                if count == max_requests:
                    return


# TODO: move code from this script to HCFPal class, leaving here only command line support
class HCFPalScript(BaseScript):

    flow_id_required = False

    def __init__(self):
        super().__init__()
        hsc = self.client._hsclient
        self.hsp = hsc.get_project(self.project_id)
        self.hcf = HCFPal(self.hsp)

    @property
    def description(self):
        return "Helper script for accessing HubCrawlFrontier."

    def add_argparser_options(self):
        super().add_argparser_options()
        subparsers = self.argparser.add_subparsers(dest="cmd")
        parser_list = subparsers.add_parser("list", help="List project frontiers or slots in a frontier")
        parser_list.add_argument("frontier", nargs="?", help="Define frontier to list it's slots")
        parser_list.add_argument("--all", action="store_true", help="List all frontiers and their slots")

        parser_count = subparsers.add_parser("count", help="Count requests in frontier slots")
        parser_count.add_argument("frontier", help="Frontier for which to count")
        parser_count.add_argument("--prefix", help="Count only slots with a given prefix", default="")
        parser_count.add_argument("--regex", help="Count only slots that matches given regex", default="")
        parser_count.add_argument(
            "--num-slots",
            type=int,
            help="Specify number of slots instead of autodetect \
                                                                 (much faster in most cases)",
        )

        parser_delete = subparsers.add_parser("delete", help="Delete slots from frontier")
        parser_delete.add_argument("frontier", help="Frontier to delete slots from")
        parser_delete.add_argument("prefix", help="Delete only slots with a given prefix")

        parser_dump = subparsers.add_parser("dump", help="Dump next requests in queue of a frontier slot")
        parser_dump.add_argument("frontier", help="Frontier name from where to dump")
        parser_dump.add_argument("slot", help="Slot from where to dump")
        parser_dump.add_argument(
            "--num-requests", help="Number of requests to dump. Defaults to %(default)d.", type=int, default=100,
        )

        parser_move = subparsers.add_parser(
            "move",
            help="Move requests from slots of given prefix, into the given \
                                                          number of slots on another prefix.",
        )
        parser_move.add_argument("frontier", help="Frontier name")
        parser_move.add_argument("prefix", help="Prefix name of the source slots")
        parser_move.add_argument("dest_prefix", help="Prefix name of the destination slots")
        parser_move.add_argument("dest_num_slots", help="Number of destination slots", type=int)
        parser_move.add_argument(
            "--num-slots",
            type=int,
            help="If given, source slots are computed using given prefix \
                                                                and this number instead of list api (sometimes list \
                                                                api works very slow)",
        )
        parser_move.add_argument(
            "--uniform",
            action="store_true",
            help="Distribute requests uniformly among slots. By default uses standard \
                                       assignation mapping.",
        )

        parser_move_batch = subparsers.add_parser(
            "move_batch", help="Move requests from given batch id into a new slot."
        )
        parser_move_batch.add_argument("frontier", help="Frontier name")
        parser_move_batch.add_argument("source_slot", help="Source slot where to find the batch id")
        parser_move_batch.add_argument("batchid", help="Id of the target batch")
        parser_move_batch.add_argument("dest_slot", help="Destination slot")
        parser_move_batch.add_argument(
            "--max-scan-batches",
            default=100,
            type=int,
            help="Max number of batches to scan in order to find target batch id in the \
                                             source slot",
        )

    def run(self):
        if self.args.cmd == "list":
            self.list_hcf()
        elif self.args.cmd == "count":
            self.count_slots()
        elif self.args.cmd == "delete":
            self.delete_slots()
        elif self.args.cmd == "dump":
            self.dump_slot()
        elif self.args.cmd == "move":
            self.move_slots()
        elif self.args.cmd == "move_batch":
            self.move_batch()
        else:
            self.argparser.print_help()

    def delete_slots(self):
        prefix_note = ' (with prefix "{}")'.format(self.args.prefix) if self.args.prefix else ""
        print(
            'Deleting slots{} from frontier "{}", project {}...'.format(
                prefix_note, self.args.frontier, self.project_id
            )
        )
        slots = [slot for slot in self.hcf.get_slots(self.args.frontier) if slot.startswith(self.args.prefix)]
        self.hcf.delete_slots(self.args.frontier, slots)
        print("Slots deleted: {}".format(slots))

    def list_hcf(self):
        if self.args.all:
            print("Listing all frontiers and their slots in project {}:".format(self.project_id))
            print(self.hcf.list_all(prettyprint=True))
        elif self.args.frontier:
            print('Listing slots for frontier "{}" in project {}:'.format(self.args.frontier, self.project_id))
            for slot in self.hcf.get_slots(self.args.frontier):
                print("\t{}".format(slot))
        else:
            print("Listing frontiers in project {}:".format(self.project_id))
            for front in self.hcf.get_frontiers():
                print("\t{}".format(front))

    def count_slots(self):
        note = ""
        if self.args.prefix:
            note = ' (with prefix "{}")'.format(self.args.prefix)
        elif self.args.regex:
            note = ' (with regex "{}")'.format(self.args.regex)
        print(
            'Counting requests in slots{} for frontier "{}", project {}:'.format(
                note, self.args.frontier, self.project_id
            )
        )
        result = self.hcf.get_slots_count(self.args.frontier, self.args.prefix, self.args.num_slots, self.args.regex)
        for slot in sorted(result["slots"].keys()):
            cnt_text = "\t{}: {}".format(slot, result["slots"][slot])
            print(cnt_text)
        print("\t" + "-" * 25)
        print("\tTotal count: {}".format(humanize.intcomma(result["total"])))
        print("\tNot-empty slots: {}".format(result["not empty slots"]))

    def dump_slot(self):
        print(
            "Dumping next {} requests from slot {}, frontier {}, pid {}:".format(
                self.args.num_requests, self.args.slot, self.args.frontier, self.project_id,
            )
        )
        for batch_id, reqs in groupby(
            self.hcf.dump_slot(self.args.frontier, self.args.slot, self.args.num_requests), key=itemgetter(0),
        ):
            print("Batch id:", batch_id)
            for _, req in reqs:
                print(req)

    def move_slots(self):
        print(
            "Moving requests from frontier {}, pid {}, prefix {} into {} slots of prefix {}".format(
                self.args.frontier, self.project_id, self.args.prefix, self.args.dest_num_slots, self.args.dest_prefix,
            )
        )
        if self.args.num_slots:
            source_slots = [self.args.prefix + str(slotno) for slotno in range(self.args.num_slots)]
        else:
            # use list api
            source_slots = [
                slot for slot in self.hcf.get_slots(self.args.frontier) if slot.startswith(self.args.prefix)
            ]
        cyclic_gen = cycle(range(self.args.dest_num_slots))
        for slot in source_slots:
            print("Reading slot %s" % slot)
            while True:
                # read each one batch
                for batch in self.hsp.frontier.read(self.args.frontier, slot, 1):
                    for fp, qdata in batch["requests"]:
                        if self.args.uniform:
                            dslotno = next(cyclic_gen)
                        else:
                            dslotno = assign_slotno(fp, self.args.dest_num_slots)
                        dslot = self.args.dest_prefix + str(dslotno)
                        self.hsp.frontier.add(self.args.frontier, dslot, [{"fp": fp, "qdata": qdata}])
                    # we don't want to generate batches bigger than source ones
                    self.hsp.frontier.flush()
                    count = len(batch["requests"])
                    self.hsp.frontier.delete(self.args.frontier, slot, [batch["id"]])
                    print("Moved batch %s (%d requests) from slot %s" % (batch["id"], count, slot))
                    break
                else:
                    break
            self.hcf.delete_slots(self.args.frontier, [slot])

    def move_batch(self):
        print(
            "Moving requests from frontier {}, pid {}, slot {}, batch {} to slot {}".format(
                self.args.frontier, self.project_id, self.args.source_slot, self.args.batchid, self.args.dest_slot,
            )
        )
        for batch in self.hsp.frontier.read(self.args.frontier, self.args.source_slot, self.args.max_scan_batches):
            if batch["id"] == self.args.batchid:
                frequests = []
                for fp, qdata in batch["requests"]:
                    frequests.append({"fp": fp, "qdata": qdata})
                self.hsp.frontier.add(self.args.frontier, self.args.dest_slot, frequests)
                self.hsp.frontier.delete(self.args.frontier, self.args.source_slot, [batch["id"]])
                self.hsp.frontier.flush()
                break


if __name__ == "__main__":
    script = HCFPalScript()
    script.run()



import hashlib
import six

try:
    from collections import Mapping
except ImportError:
    from collections.abc import Mapping

from scrapinghub.client import parse_auth


def convert_from_bytes(data):
    if data is not None:
        data_type = type(data)
        if data_type == bytes:
            return data.decode('utf8')
        if data_type in (str, int, float, bool):
            return data
        if isinstance(data, Mapping):
            # Includes dict, OrderedDict, UserDict, scrapy.Item...
            data = data.items()
        return data_type(map(convert_from_bytes, data))


def convert_to_bytes(data):
    if data is not None:
        data_type = type(data)
        if data_type == str:
            return data.encode('utf8')
        if data_type in (bytes, int, float, bool):
            return data
        if data_type == dict:
            data = data.items()
        return data_type(map(convert_to_bytes, data))


def hash_mod(text, divisor):
    """
    returns the module of dividing text md5 hash over given divisor
    """
    if isinstance(text, six.text_type):
        text = text.encode('utf8')
    md5 = hashlib.md5()
    md5.update(text)
    digest = md5.hexdigest()
    return int(digest, 16) % divisor


def assign_slotno(path, numslots):
    """
    Standard way to assign slot number from url path
    """
    return str(hash_mod(path, numslots))


def get_apikey():
    """Provides a facade to the multiple behaviors of parse_auth() and forces
    it to read the 'SH_APIKEY' env var and return it.
    """

    try:
        return parse_auth(None)[0]
    except RuntimeError:
        pass



import logging
import time
from collections import defaultdict

import requests as requests_lib
from scrapinghub import ScrapinghubClient

from hcf_backend.utils import get_apikey


LOG = logging.getLogger(__name__)


class HCFManager(object):
    def __init__(self, project_id, frontier, auth=None, batch_size=0):
        if auth is None:
            auth = get_apikey()
        self._client = ScrapinghubClient(auth=auth)
        self._hcf = self._client.get_project(project_id).frontiers
        self._frontier = self._hcf.get(frontier)
        self._links_count = defaultdict(int)
        self._links_to_flush_count = defaultdict(int)
        self._batch_size = batch_size
        self._hcf_retries = 10

    def add_request(self, slot, request):
        self._frontier.get(slot).q.add([request])
        self._links_count[slot] += 1
        self._links_to_flush_count[slot] += 1
        if self._batch_size and self._links_to_flush_count[slot] >= self._batch_size:
            return self.flush(slot)
        return 0

    def flush(self, slot=None):
        n_links_to_flush = self.get_number_of_links_to_flush(slot)
        if n_links_to_flush:
            if slot is None:
                self._hcf.flush()
                for slot in self._links_to_flush_count.keys():
                    self._links_to_flush_count[slot] = 0
                LOG.info("Flushed %d link(s).", n_links_to_flush)
            else:
                slot_obj = self._frontier.get(slot)
                slot_obj.flush()
                self._links_to_flush_count[slot] = 0

        return n_links_to_flush

    def read(self, slot, mincount=None):
        slot_obj = self._frontier.get(slot)
        for i in range(self._hcf_retries):
            try:
                return slot_obj.q.iter(mincount=mincount)
            except requests_lib.exceptions.ReadTimeout:
                LOG.error(
                    "Could not read from {0}/{1} try {2}/{3}".format(self._frontier.key, slot, i + 1, self._hcf_retries)
                )
            except requests_lib.exceptions.ConnectionError:
                LOG.error(
                    "Connection error while reading from {0}/{1} try {2}/{3}".format(
                        self._frontier.key, slot, i + 1, self._hcf_retries
                    )
                )
            except requests_lib.exceptions.RequestException:
                LOG.error(
                    "Error while reading from {0}/{1} try {2}/{3}".format(
                        self._frontier.key, slot, i + 1, self._hcf_retries
                    )
                )
            time.sleep(60 * (i + 1))
        return []

    def delete(self, slot, ids):
        slot_obj = self._frontier.get(slot)

        for i in range(self._hcf_retries):
            try:
                slot_obj.q.delete(ids)
                break
            except requests_lib.exceptions.ReadTimeout:
                LOG.error(
                    "Could not delete ids from {0}/{1} try {2}/{3}".format(
                        self._frontier.key, slot, i + 1, self._hcf_retries
                    )
                )
            except requests_lib.exceptions.ConnectionError:
                LOG.error(
                    "Connection error while deleting ids from {0}/{1} try {2}/{3}".format(
                        self._frontier.key, slot, i + 1, self._hcf_retries
                    )
                )
            except requests_lib.exceptions.RequestException:
                LOG.error(
                    "Error deleting ids from {0}/{1} try {2}/{3}".format(
                        self._frontier.key, slot, i + 1, self._hcf_retries
                    )
                )
            time.sleep(60 * (i + 1))

    def delete_slot(self, slot):
        slot_obj = self._frontier.get(slot)
        slot_obj.delete()

    def close(self):
        self._hcf.close()
        self._client.close()

    def get_number_of_links(self, slot=None):
        if slot is None:
            return sum(self._links_count.values())
        else:
            return self._links_count[slot]

    def get_number_of_links_to_flush(self, slot=None):
        if slot is None:
            return sum(self._links_to_flush_count.values())
        else:
            return self._links_to_flush_count[slot]



"""
shub-workflow based script (runs on Zyte ScrapyCloud) for easy managing of consumer spiders from multiple slots. It checks available slots and schedules
a job for each one.
"""

import json
import random
import logging

import humanize
from shub_workflow.crawl import CrawlManager

from hcf_backend.utils.hcfpal import HCFPal


logging.basicConfig(format="%(message)s", level=logging.INFO)
logger = logging.getLogger(__name__)


class HCFCrawlManager(CrawlManager):

    default_max_jobs = 1

    def __init__(self):
        super().__init__()
        self.hcfpal = HCFPal(self.client._hsclient.get_project(self.project_id))

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument("frontier", help="Frontier name")
        self.argparser.add_argument("prefix", help="Slot prefix")
        self.argparser.add_argument("--frontera-settings-json")

    @property
    def description(self):
        return __doc__

    def print_frontier_status(self):
        result = self.hcfpal.get_slots_count(self.args.frontier, self.args.prefix)
        logger.info(f"Frontier '{self.args.frontier}' status:")
        for slot in sorted(result["slots"].keys()):
            cnt_text = "\t{}: {}".format(slot, result["slots"][slot])
            logger.info(cnt_text)
        logger.info("\tTotal count: {}".format(humanize.intcomma(result["total"])))

        return set(result["slots"].keys())

    def workflow_loop(self):
        available_slots = self.print_frontier_status()

        running_jobs = 0
        states = "running", "pending"
        for state in states:
            for job in self.get_project().jobs.list(
                spider=self.args.spider, state=state, meta="spider_args"
            ):
                frontera_settings_json = json.loads(
                    job["spider_args"].get("frontera_settings_json", "{}")
                )
                if "HCF_CONSUMER_SLOT" in frontera_settings_json:
                    slot = frontera_settings_json["HCF_CONSUMER_SLOT"]
                    if slot in available_slots:
                        available_slots.discard(slot)
                        running_jobs += 1

        available_slots = [
            slot
            for slot in available_slots
            if self.hcfpal.get_slot_count(self.args.frontier, slot) > 0
        ]
        logger.info(f"Available slots: {available_slots!r}")
        if available_slots:
            random.shuffle(available_slots)
            if self.max_running_jobs:
                max_jobs = self.max_running_jobs - running_jobs
                available_slots = available_slots[:max_jobs]
                if not available_slots:
                    logger.info("Already running max number of jobs.")

            base_frontera_settings = {}
            if self.args.frontera_settings_json:
                base_frontera_settings = json.loads(self.args.frontera_settings_json)
            for slot in available_slots:
                frontera_settings = base_frontera_settings.copy()
                frontera_settings.update(
                    {
                        "HCF_CONSUMER_SLOT": slot,
                        "HCF_CONSUMER_FRONTIER": self.args.frontier,
                    }
                )
                frontera_settings_json = json.dumps(frontera_settings)
                logger.info(
                    f"Will schedule spider job with frontera settings {frontera_settings_json}"
                )
                jobkey = self.schedule_spider(
                    spider_args_override={
                        "frontera_settings_json": frontera_settings_json
                    }
                )
                logger.info(
                    f"Scheduled job {jobkey} with frontera settings {frontera_settings_json}"
                )
            return True
        return bool(running_jobs)


if __name__ == "__main__":
    manager = HCFCrawlManager()
    manager.run()



# Automatically created by: shub deploy

from setuptools import setup, find_packages

setup(
    name         = 'hcf-backend',
    version      = '0.4.6',
    description  = 'ScrapyCloud HubStorage frontier backend for Frontera',
    long_description = open('README.md').read(),
    long_description_content_type = 'text/markdown',
    license      = 'BSD',
    url          = 'https://github.com/scrapinghub/hcf-backend',
    maintainer   ='Scrapinghub',
    packages     = find_packages(),
    install_requires = (
        'frontera==0.7.1',
        'humanize==0.5.1',
        'requests>=2.18.4',
        'retrying>=1.3.3',
        'scrapinghub>=2.0.0',
        'shub-workflow>=1.5.1',
    ),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)
